//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the concrete material interface.
//				
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_CONCRETE_MATERIAL_HPP
#define OC_I_CONCRETE_MATERIAL_HPP

#include <mxf_tk/I_generic_material.hpp>
#include <mxf_tk/I_timecode.hpp>
#include <mxf_tk/I_concrete_track.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /* wrapping:
     * std_wrapping selects the default wrapping (frame or clip wrapping depending on the essence)
     */
    enum wrapping
    {
      // standard wrappings
      std_wrapping, 
      clip_wrapping, 
      frame_wrapping, 
      line_wrapping, 
      custom_wrapping, 
      // manufacturer wrappings
      evtr_wrapping, 
      xdcam_wrapping, 
      p2_wrapping, 
      k2_wrapping,
      avid_opatom_wrapping,
      // special
      opzero_wrapping, 
      dcp_wrapping,  // SMPTE
      dcp1_wrapping, // Interop
      // error
      wrapping_error
    };

    /* I_concrete_material:
     * A class describing a material with an embedded essence
     * In mxf terms: Top Level Source Package.
     */
    class I_concrete_material : public I_generic_material
    {
    public:
      /**
       * This function will operate in streaming mode only!
       * Concrete material was created using NewStreamMaterialInterface! 
       * Append content of the const uint8_t* buffer to the 
       * current concrete material. The size_type parameter 
       * indicates the size of the buffer (in bytes).
       * The integer is a stream_id allowing you to specify to
       * which stream the uint8_t* buffer should be appended.
       * Returns the amount of data that could be effectively written. Failure
       * to write all the data usually means that the internal buffer is full and that
       * you are writing faster than MXFTk can read during a flush operation. Another 
       * reason can be that you try to write more than the actual size of the buffer.
       */
      virtual size_t CALLING_CONVENTION write(unsigned int /* in */,
                                              const uint8_t* /* in free */, 
                                              size_t /* in */)=0;

      /**
       * This function will operate in streaming mode only!
       * This function must be called once all the essence data
       * of the stream has been written. 
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION end_of_stream(unsigned int /* in */)=0;

      /*
       * Returns the wrapping of the essence.
       */
      virtual wrapping CALLING_CONVENTION get_wrapping() const=0;

      /*
       * Computes the duration of this concrete material.
       * After calling this function, the tracks will have a correct duration.
       * You can use this function when wrapping MXF files and when you need to
       * know the duration before flushing the MXF file.
       * This function must be called before attaching the material to an instance of I_mxf_file.
       * This function will be inactive when the input is a stream or while reading MXF files.
       */
      virtual void CALLING_CONVENTION compute_duration()=0;
    };

    /*
     * Retrieves a concrete material interface.
     * I_concrete_material** pointer to the newly allocated object
     *
     * locators: list of source files (DV, MPEG, etc.)
     * In order to designate a series of images the following syntax should be 
     * set in the path of the locator #[d, f, l, s] where:
     * ->d should be a number designating the number of digits
     * ->f should be the first number of the file to be wrapped
     * ->l should be the last number of the file to be wrapped
     * ->s should be the step between two successive files to be wrapped
     * Note that it is perfectly valid to set a negative step.
     * For instance 
     * "C:\\images\\myfile#[4, 10, 500, 2].jp2", "C:\\sound\\sound.wav"
     * will perform the wrapping of the following series of files
     * C:\\images\\myfile0010.jp2
     * C:\\images\\myfile0012.jp2
     * C:\\images\\myfile0014.jp2
     * ...
     * C:\\images\\myfile0498.jp2
     * C:\\images\\myfile0500.jp2
     * with the audio wave "sound.wav" 
     *
     * "C:\\video\\test.dv", "C:\\sound\\sound.wav"
     * will perform the wrapping of "test.dv" with "sound.wav"
     *
     * wrapping: desired wrapping of the essence.
     *
     * rational*: set the default edit rate value to be used when this information
     * cannot be retrieved from the essence (e.g. Uncompressed Images).
     * If the last boolean is set to true DV sources with embedded audio will 
     * be considered as video only.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewConcreteMaterialInterface(I_concrete_material** /* out free */, 
                                                                               locators* /* in free */, 
                                                                               wrapping = std_wrapping /* in */, 
                                                                               rational* r = NULL /* in free */, 
                                                                               bool ignore_audio_from_dv = false /* in */);

    
    /*
     * Retrieves a concrete material interface that will point on an externally referenced media file.
     * I_concrete_material** pointer to the newly allocated object
     *
     * const char* : path to the external media file to be referenced.
     *
     * rational*: set the default edit rate value to be used when this information
     * cannot be retrieved from the essence (e.g. Uncompressed Images).
     * If the last boolean is set to true DV sources with embedded audio will 
     * be considered as video only.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewExtConcreteMaterialInterface(I_concrete_material** /* out free */, 
                                                                                  const char* /* in free */, 
                                                                                  rational* r = NULL /* in free */, 
                                                                                  bool ignore_audio_from_dv = false /* in */);

    /*
     * Retrieves a concrete material interface.
     * function is the same as previous one but uses crypted_locators instead of locators
     *
     * wrapping: desired wrapping of the essence (std_wrapping, frame_frapping, opzero_wrapping,
     * dcp_wrapping or dcp1_wrapping).
     *
     * If the last boolean is set to true DV sources with embedded audio will 
     * be considered as video only.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewCryptedMaterialInterface(I_concrete_material** /* out free */, 
                                                                              crypted_locators* /* in free */, 
                                                                              wrapping = frame_wrapping /* in */, 
                                                                              rational* r = NULL /* in free */, 
                                                                              bool ignore_audio_from_dv = false /* in */);
  
    /*
     * Retrieves a stream material interface.
     * I_stream_material** pointer to the newly allocated object
     * wrapping desired wrapping of the essence in this material.
     * rational*: set the default edit rate value to be used when this information
     * cannot be retrieved from the essence (e.g. Uncompressed Images).
     * If the last boolean is set to true DV sources with embedded audio will 
     * be considered as video only.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewStreamMaterialInterface(I_concrete_material** /* out free */, 
                                                                             I_essence_stream_task* /* in free */, 
                                                                             wrapping = std_wrapping /* in */, 
                                                                             rational* r = NULL /* in free */, 
                                                                             bool ignore_audio_from_dv = false /* in */);

    /*
     * Frees a concrete material interface.
     * I_concrete_material** pointer to the object to be freed
     * User must not free a concrete material already attached to an I_mxf_file.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeConcreteMaterialInterface(I_concrete_material** /* in out */);
  }
}

#endif
