//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK : Error handler.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_MXF_ERROR_HPP
#define OC_I_MXF_ERROR_HPP

#include <mxf_tk/inttypes.hpp>
#include <mxf_tk/mxf_tk.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
  
    class error_list;

    /*
     * This class reflects the seriousness of the error.
     * It can be generally assumed that a
     * * caution error simply indicates an unexpected bahavior or data 
     *   but doesn't stop the API from running correctly.
     * * warning error indicates an abnormal behavior or an invalid
     *   data. The data manipulated might be corrupted.
     * * fatal error indicates that the API is not able to continue
     *   to run correctly.
     */
    typedef enum {  MXF_NONE = uint32_t(0),
		    MXF_CAUTION,
		    MXF_WARNING,
		    MXF_FATAL  } gravity;

    /*
     * An interface to handle API internal error.
     * Includes support for internationalization.
     */
    class I_mxf_error
    {
    public:
    
      /*
       * Returns the id of the error
       */
      virtual uint32_t CALLING_CONVENTION get_id() const=0;

      /*
       * Returns the error gravity (from fatal to caution) 
       */
      virtual gravity CALLING_CONVENTION get_gravity() const=0;

      /*
       * Returns the unicode string describing the error
       */
      virtual const wchar_t* CALLING_CONVENTION get_error_msg() const=0;

      /*
      * Returns the id of the thread where the error occured.
      */
      virtual uint64_t CALLING_CONVENTION get_thread_id() const=0;
    };

    /*
     * An Interface to handle MxtTk errors.
     */
    class I_mxf_error_handler
    {
    public:
      /*
       * Set the error messages dictionary file.
       * Returns false if an error occured.
       */
      virtual bool CALLING_CONVENTION set_error_dictionary(const char* /* in free */)=0;

      /*
       * Returns true if an error is to read!
       */
      virtual bool CALLING_CONVENTION has_an_error() const=0;

      /*
       * Returns the list of errors.
       * Returns NULL if an error occured while retrieving the list of errors
       */
      virtual error_list* /* free */ CALLING_CONVENTION get_errors()=0;

      /*
       * Frees the list of errors returned by get_errors().
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION free_errors(error_list* /* in */) const=0;

      /*
       * Clear the list of errors.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION clear()=0;

      /*
       * Clear the list of errors coming from the given thread id
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION clear_thread_error(uint64_t)=0;
    };

    /*
     * Retrieves the I_mxf_error_handler.   
     * Default language is set to English - United States. (dictionary filename: mxf-en-US.dic)
     * Use the fonction SetLanguage to change the error message language ( the corresponding error message dictionary should exist).
     * The user doesn't need to Free this interface.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION GetMxfErrorHandlerInterface(I_mxf_error_handler** /* out */);
  }
}
      
#endif
