//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the mxf file interface.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_MXF_FILE_HPP
#define OC_I_MXF_FILE_HPP

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_metadata.hpp>
#include <mxf_tk/I_generic_material.hpp>
#include <mxf_tk/I_concrete_material.hpp>
#include <mxf_tk/I_concrete_track.hpp>
#include <mxf_tk/streaming.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /*
     * Parameters of the mxf_file.
     * KAG_SIZE sets the kag size
     * HEADER_REPETITION indicates how often the header metadata should
     * be repeated (for instance "2" would mean every two partitions. The
     * default value of 0 will perform a repetition only when the header values
     * change.
     * INDEX_TABLE activates or deactivates the index table generation (when posible).
     * PREFERRED_PARTITION_DURATION sets the desired duration (in seconds) of each partition
     * if set to "0.0" there won't be any body partitions.
     * REVERSE PLAY activated the creation of reverse play klvs.
     */
    enum { KAG_SIZE, HEADER_REPETITION, INDEX_TABLE, PREFERRED_PARTITION_DURATION, REVERSE_PLAY };

    /*
     * MXF opening mode is a parameter used to optimize the
     * opening of an MXF file depending on what you want to do with this file
     * You must set this parameter when calling NewMxfFileInterface()
     * METADATA_ONLY : Fastest opening mode but you will only get the first
     *                 header metadata found while decoding the file. Depending 
     *                 on the file, this header metadata may be open and/or incomplete.
     *                 The metadata you get should be taken with "caution".
     *                 You cannot extract media from the file with this mode.
     * CLOSED_METADATA_ONLY : Return the best closed header metadata found
     *                 in the file. The header metadata can be incomplete, that
     *                 means all the metadata you get will be valid but some of it
     *                 might be missing.
     * CLOSED_COMPLETE_METADATA : Return the best header metadata found
     *                 in the file. This process may be longer than CLOSED_METADATA_ONLY
     *                 but you are sure to get valid metadata.
     *                 You cannot extract media from the file with this mode.
     * METADATA_AND_LINEAR_PLAYOUT: Similar to CLOSED_AND_COMPLETE_METADATA_ONLY
     *                 but also enable extraction of media stored in the file.
     * METADATA_AND_RANDOM_ACCESS: Similar to METADATA_AND_LINEAR_PLAYOUT and 
     *                 adds decoding of index tables enabling faster random access
     *                 to the media.
     */
    enum mxf_opening_mode { METADATA_ONLY, 
                            CLOSED_METADATA_ONLY,
                            CLOSED_COMPLETE_METADATA,
                            METADATA_AND_LINEAR_PLAYOUT,
                            METADATA_AND_RANDOM_ACCESS };

    /* I_mxf_file:
     * An interface to read or edit an mxf_file already existing.
     */
    class I_mxf_file
    {
    public:
      /**
       * This function will operate in mxf input streaming mode only!
       * mxf_file was created using NewMxfStreamInterface! 
       * Append content of the const uint8_t* buffer to the 
       * current mxf stream. The size_type parameter 
       * indicates the size of the buffer (in bytes).
       * Return the amount of data that could be effectively written. Failure
       * to write all usually means that the internal buffer is full and that
       * you are writing faster than MxfTk can read during the file decoding. 
       * Another reason can be that you try to add more than the actual size 
       * of the buffer.
       */
      virtual size_t CALLING_CONVENTION write(const uint8_t* /* in free */, 
                                              size_t /* in */)=0;

      /**
       * This function will operate in mxf input streaming mode only!
       * This function must be called once the mxf stream is closed.
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION end_of_stream()=0;

      /*
       * Returns the operational pattern of the file. Returned pointer
       * mus not be destroyed.
       * Returns NULL if an error occured.
       */
      virtual const char* CALLING_CONVENTION operational_pattern() const=0;

      /*
       * Returns true if the file contains one or more 
       * external essences.
       */
      virtual bool CALLING_CONVENTION has_external_essence() const=0;

      /*
       * Returns a list of output materials in this file.
       * Gives information on file playout by referencing
       * data from embedded materials.
       * In MXF terms (Material Package)
       * Return NULL if an error occured.
       */
      virtual generic_list* /* free */ CALLING_CONVENTION output_material()=0;

      /*
       * Returns the embedded materials of this file.
       * These materials contain essence data to be read.
       * In MXF terms (Top Level Source Package)
       * Returns NULL if an error occured
       */
      virtual concrete_list* /* free */ CALLING_CONVENTION embedded_material()=0;

      /*
       * Returns the low level materials of this file (usually none)
       * In MXF terms (Low Level Source Package)
       * These packages are provided for expert end-users and do
       * not contain essence data.
       * Returns NULL if an error occured.
       */
      virtual generic_list* /* free */ CALLING_CONVENTION low_level_material()=0;

      /*
       * Frees the list returned by output_material or low_level_material
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION free_generic_list(generic_list* /* in */)=0;

      /*
       * Frees the list returned by embedded_material
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION free_concrete_list(concrete_list* /* in */)=0;

      /*
       * NOTE TO PREVIOUS VERSIONS USERS:
       * This function is now threaded! Read carefully what is written below.
       * If this I_mxf_file was opened for writing, the call to this function will cause
       * the creation of the mxf file (on the disk or on a stream).
       * A thread will be created so the flush will not be completed when returning for this function. 
       * You should call mxf_file::wait_end_flush_thread and wait for its termination.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION flush()=0;

      /*
       * Creates the mxf file.
       * This flush lets you specify the name of the mxf file being generated. 
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION flush_file(const char* /* in free */)=0;

      /*
       * Creates the mxf file.
       * This flush lets you specify the streaming task that will handle the output stream.
       * MxfTk user should provide its own implemetation of I_output_mxf_stream_task.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION flush_stream(I_output_mxf_stream_task* /* in free */)=0;

      /*
       * Calls this function to know if the flush thread is still running.
       * You should not free the I_mxf_file interface as long as the
       * process is not completed.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION flush_thread_running()=0;

      /*
       * This function will return when the flush thread has finished.
       * Returns false if the process was cancelled, true otherwise. 
       */
      virtual bool CALLING_CONVENTION wait_end_flush_thread()=0;

      /*
       * This function returns the current progress of the process:
       * 0.0 -> beginning / 1.0 -> end. / <0 -> progress cannot be computed.
       */
      virtual double CALLING_CONVENTION flush_thread_progress()=0;

      /*
       * Calls this function to cancel the flush thread currently running.
       * The cancellation is asynchronous so you must still call wait_end_flush_thread()
       * to terminate properly the process.
       */
      virtual bool CALLING_CONVENTION flush_thread_cancel()=0;

      /*
       * Call this function to retrieve the id of the flush thread
       */
      virtual uint64_t CALLING_CONVENTION flush_thread_id()=0;

      /*
       * Returns the preface set of this file (header metadata root).
       * Returns NULL if an error occured.
       */
      virtual const I_metadata* CALLING_CONVENTION get_preface() const=0;

      /*
       * Returns the identifications set of this file.
       * Provides an history of who edited when this file.
       * Returns NULL if an error occured.
       */
      virtual metadata_list* /* free */ CALLING_CONVENTION get_identifications() const=0;

      /*
       * Frees the metadata_list returned by get_identifications
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION free_identifications(metadata_list* /* in */) const=0;

      /*
       * Returns one of the file's parameter value.
       * This function will return the string "unknown" if this parameter 
       * can not be retrieved.
       * Returns NULL if an error occured.
       */
      virtual const char* CALLING_CONVENTION get_parameter(int) const=0;

      /*
       * Sets a parameter value.
       * int is KAG_SIZE, HEADER_REPETITION, INDEX_TABLE
       * PREFERRED_PARTITION_DURATION or REVERSE_PLAY
       * KAG_SIZE value is an integer giving the size in bytes; for instance "512" (default "512").
       * This can be used to align KLV items on a spacing grid.
       * INDEX_TABLE value is "True" or "False" (default "False"). Please be
       * aware that some essences always require an index table (D10 for instance).
       * HEADER_REPETITION indicates how often the header metadata should
       * be repeated (for instance "2" would mean every two partitions. The
       * default value of "0" will perform a repetition only if the header has
       * changed (default "0").
       * PREFERRED_PARTITION_DURATION sets the desired duration of each partition
       * if set to "0" there won't be any body partitions (default "0").
       * Return false if parameter set couldn't be performed.
       * REVERSE_PLAY indicates reverse play elements are expected.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION set_parameter(int /* in */, 
                                                    const char* /* in free */)=0;

      /*
       * Outputs the entire header metadata in an xml file.
       * This is convenient for fast analysis of the file and debugging.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION output_xml(const char* /* in free */)=0;

      /*
       * Some MXF files (usually SONY XDCAM files) include an XML file within 
       * the header metadata. This function will write a file on disk if such
       * an XML content is found.
       */
      virtual bool CALLING_CONVENTION output_embedded_xml(const char* /* in free */)=0;

      /*
       * Output the entire header metadata in an xml buffer.
       * This is convenient for fast analysis of the file and debugging.
       * Returns pointer on buffer if successful.
       */
      virtual const unsigned char * CALLING_CONVENTION output_xml_buffer(uint64_t & /* out */) = 0;

      /*
       * Get stream identifier of waiting stream in active streaming mode.
       * Parameter 1 is set to the waiting stream task, parameter 2 is set 
       * to the witing stream id of this stream task.
       * Returns false if no stream is waiting or if mode is different.
       */
      virtual bool CALLING_CONVENTION waiting_stream(I_essence_stream_task*& /* out */, 
                                                     unsigned int & /* out */) = 0;
    };

    /*
     * Retrieves an I_mxf_file interface.
     * I_mxf_file** pointer to the newly allocated object
     * const char* points to the mxf file to be read.
     * The third parameter (update_upon_closing) should be set to true if you wish to update
     * the file (change timecode, umid, descriptive metadata). In that case, the file will be
     * updated on disk upon destruction of the I_mxf_file interface. 
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewMxfFileInterface(I_mxf_file** /* out free */, 
                                                                      const char* /* in free */, 
                                                                      mxf_opening_mode = METADATA_AND_RANDOM_ACCESS /* in */);

    /*
     * Retrieves an I_mxf_file interface.
     * I_mxf_file** pointer to the newly allocated object
     * I_input_mxf_stream_task* pointer to the user's class that will feed this stream.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewMxfStreamInterface(I_mxf_file** /* out free */, 
                                                                        I_input_mxf_stream_task* /* in free */);

    /*
     * Retrieves a partial I_mxf_file interface.
     * I_mxf_file** pointer to the newly allocated object
     * const char* points to the mxf file to be read.
     * The timecode values correpond to the timecode in and timecode out to be retrieved.
     * Timecode pointers should be destroyed by the API user.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewPartialRestoreMxfFileInterface(I_mxf_file** /* out free */, 
                                                                                    const char* /* in free */, 
                                                                                    I_timecode* /* in free */, 
                                                                                    I_timecode* /* in free */);

    /*
     * Retrieves an I_mxf_file interface.
     * I_mxf_file** pointer to the newly allocated object
     * I_input_mxf_stream_task* pointer to the user's class that will feed this stream.
     * The timecode values correpond to the timecode in and timecode out to be retrieved.
     * Timecode pointers should be destroyed by the API user.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewPartialRestoreMxfStreamInterface(I_mxf_file** /* out free */, 
                                                                                      I_input_partial_mxf_stream_task* /* in free */, 
                                                                                      I_timecode* /* in free */, 
                                                                                      I_timecode* /* in free */);

    /*
     * Free an mxf_file interface.
     * I_mxf_file** pointer to the object to be freed.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeMxfFileInterface(I_mxf_file** /* in */);

    /*
     * Frees an mxf_file interface and update it on disk if it was opened 
     * for reading. Calling this function will flush the changes you made to the
     * file (timecode, descriptive metadata, etc...)
     * In that case, the mxf stream/file  will be updated . If you ar using the streaming interface 
     * you must define I_input_mxf_stream_task as seekable and provide an implementation of
     * the function I_input_mxf_stream_task::updated_mxf_file().
     * I_mxf_file** pointer to the object to be freed.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION UpdateAndFreeMxfFileInterface(I_mxf_file** /* in */);
  }
}

#endif
