//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the op2b file interface.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_OP2B_FILE_HPP
#define OC_I_OP2B_FILE_HPP

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_mxf_file.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /* I_op2b_file:
     * An interface to create an op2b mxf file.
     */
    class I_op2b_file : public I_mxf_file
    {
    private:
      virtual size_t CALLING_CONVENTION write(const uint8_t*, size_t)=0;
      virtual bool CALLING_CONVENTION end_of_stream()=0;
    public:
      /*
       * Notifies MXFTk that the next I_concrete_material added to this file will be multiplexed
       * For instance to create the following generic material (a,b,c,d,e,f designate 
       * concrete material) :
       * track 1 : a+b+c
       * track 2 : d+e+f
       * you should do the following calls:
       * new_multiplexed_material()
       * add_material(a)
       * add_material(d)
       * new_multiplexed_material()
       * add_material(b)
       * add_material(e)
       * new_multiplexed_material()
       * add_material(c)
       * add_material(f)
       *
       * Note that each multiplexed concrete material should have the same duration.
       */
      virtual bool CALLING_CONVENTION new_multiplexed_material()	=0;

      /*
       * Set a material to be wrapped in the mxf_file.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION add_material(I_concrete_material* /* in */)	=0;

       /*
       * Set an external material to be wrapped in the mxf_file.
       * The second parameter is used to set the locator string (usually the path
       * to the external file) that will be included in the header metadata.
       * Return true if successful.
       */
      virtual bool CALLING_CONVENTION add_external_material(I_concrete_material* /* in */, 
                                                            const wchar_t* /* in free */)=0;

      /*
       * Call this function to state whether special processing is required at the join of two 
       * source clips of a track of the genric material. This value is used by the MXF decoder
       * to know if it can play continuously the essences. 
       * For instance if the file will contain a DV and MPEG in the same track, you should
       * set this value to false (default). Whenever special processing is required to
       * play at least one of the generic material's track, this value should be set to false.
       */
      virtual bool CALLING_CONVENTION set_continuous_decoding(bool)=0;
    
      /*
       * Return the generic material created after calling set_material or set_external_material
       * Return NULL if the generic material was not created yet.
       * Do not call add_material or add_external_material after calling this function
       * or all the changes you made to the generic material will be lost.
       */
      virtual I_generic_material* CALLING_CONVENTION get_current_generic_material()=0;
    };

    /*
     * Retrieves an I_op1b_file interface.
     * I_op2b_file** pointer to the newly allocated object
     * const char* path to the mxf file to be written.
     * Call to flush() will effectively cause the on-disk writing of
     * the file.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewOp2bFileInterface(I_op2b_file** /* out free */, 
                                                                       const char* /* in free */);

    /*
     * Retrieves an I_op2b_file interface. The flush will be performed in streaming mode.
     * I_op2b_file** pointer to the newly allocated object
     * const char* path to the mxf file to be written.
     * Call to flush() will effectively cause the on-disk writing of
     * the file.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewOp2bStreamInterface(I_op2b_file** /* out free */, 
                                                                         I_output_mxf_stream_task* /* in free */);

    /*
     * Frees an op2b_file interface.
     * I_op2b_file** pointer to the object to be freed.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeOp2bFileInterface(I_op2b_file** /* in out */);
  }
}

#endif
