//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the metadata material interface.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_METADATA_MATERIAL_HPP
#define OC_I_METADATA_MATERIAL_HPP

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_generic_material.hpp>
#include <mxf_tk/I_metadata.hpp>
#include <mxf_tk/I_timecode.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /* 
     * Descriptive metadata types:
     * m_timeline: A timeline metadata defines a track where metadata
     *             is continuously appended - at any point in the track metadata MUST be found).
     *             The metadata is always ordered in a time-linear fashion and metadata segments
     *             cannot overlap.
     * m_event: An event metadata defines a track where metadata is
     *          not continuously appended - there might be metadata segments overlapping on 
     *          this track and some sections of the track may not contain metadata at all.
     * m_static: A static metadata defines a track without any time reference (origin and duration
     *           are irrelevant for this track). It is used to set metadata global to the
     *           entire material.
     */
    enum metadata_type {m_timeline, m_event, m_static, m_error};

    /* I_metadata_material: 
     * Material encapsulating a descriptive metadata framework as well as time and track
     * references.
     */
    class I_metadata_material : public I_generic_material
    {
    private:
      /*
       * Disabled I_generic_material function.
       */
      virtual I_track* CALLING_CONVENTION new_metadata_track(track_type, int64_t, rational*)=0;
      virtual I_track* CALLING_CONVENTION new_stream_metadata_track(I_input_metadata_stream_task*, track_type, int64_t, rational*)=0;
      virtual track_list* CALLING_CONVENTION source_tracks()=0;
      virtual track_list* CALLING_CONVENTION metadata_tracks()=0;
      virtual bool CALLING_CONVENTION free_tracks(track_list*)=0;
      virtual I_timecode_material* CALLING_CONVENTION get_timecode() const=0;
      virtual bool CALLING_CONVENTION set_timecode(I_timecode_material*)=0;
      virtual bool CALLING_CONVENTION set_material_name(const wchar_t*, size_t)=0;
      virtual const wchar_t* CALLING_CONVENTION get_material_name(size_t&) const=0;
      virtual bool CALLING_CONVENTION set_umid(const I_umid*)=0;
      virtual I_umid* CALLING_CONVENTION get_umid()=0;
    public:
      /*
       * Returns the metadata type
       * Returns m_error if an error occured.
       */
      virtual metadata_type CALLING_CONVENTION get_metadata_type() const=0;

      /*
       * Returns the metadata tree (usually root of a framework).
       * Returned I_metadata object can be edited, it must not
       * be deleted.
       * Returns NULL if an error occured.
       */
      virtual I_metadata* CALLING_CONVENTION get_metadata() const=0;

      /*
       * Replaces the current descriptive metadata tree (and deletes it) with
       * the new one.
       */
      virtual bool CALLING_CONVENTION replace_metadata(I_metadata* /* in */)=0;


      /*
       * Offset into the descriptive metadata track to which this material
       * will be appended. Measured in edit units of the track.
       * Returns 0x7fffffffffffffff if an error occured.
       */
      virtual int64_t CALLING_CONVENTION get_start_position() const=0;

      /*
       * Duration into the descriptive metadata track to which this material
       * will be appended. Measured in edit units of the track.
       * Returns 0x7fffffffffffffff if an error occured
       */
      virtual int64_t CALLING_CONVENTION get_duration() const=0;

      /*
       * Returns the utf16 string comment of this metadata. The size_t parameter
       * will be returned with the length (in bytes) of the comment. Returned
       * pointer can be NULL if not comment is provided.
       * Returns NULL if an error occured.
       * Returned pointer should not be freed.
       */
      virtual wchar_t* CALLING_CONVENTION get_comment(size_t&) const=0;
    };

    /*
     * Retrieves a metadata material interface.
     * I_metadata_material** pointer to the newly allocated object
     * metadata_type type of metadata to be created
     * int64_t start position (in edit units) in the track where this material will be added.
     *	+ will be ignored if metadata_type is m_timeline or m_static.
     * int64_t duration (in edit units) in the track where this material will be added.
     *	+ will be ignored if metadata_type is m_static.
     * wchar_t* utf16 string comment for this metadata material
     * size_t length in bytes of the comment.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewMetadataMaterialInterface(I_metadata_material** /* out free */, 
                                                                               metadata_type /* in */, 
                                                                               I_metadata* /* in */, 
                                                                               int64_t /* in */, 
                                                                               int64_t /* in */, 
                                                                               const wchar_t* /* in free */, 
                                                                               size_t /* in */);

    /*
     * Frees a metadata material interface.
     * I_metadata_material** pointer to the object to be freed
     * User should not free a material already appended to an I_track.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeMetadataMaterialInterface(I_metadata_material** /* in out */);
  }
}

#endif
