//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the op3c file interface.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_OP3C_FILE_HPP
#define OC_I_OP3C_FILE_HPP

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_mxf_file.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /* I_op3c_file:
     * An interface to create an op3c mxf file.
     */
    class I_op3c_file : public I_mxf_file
    {
    private:
      virtual size_t CALLING_CONVENTION write(const uint8_t*, size_t)=0;
      virtual bool CALLING_CONVENTION end_of_stream()=0;
    public:
       /*
       * This function creates a new generic material (Material Package)
       * The following calls to new_generic_material_material(), add_material_track() and add_external_material_track()
       * will apply to the newly created generic material.
       */
      virtual bool CALLING_CONVENTION new_generic_material()=0;

      /*
       * Notifies MXFTk that the next concrete tracks added to this file will be part 
       * of a new genric material track
       * For instance to create the following generic material (a,b,c,d,e designate 
       * concrete tracks) :
       * track 1 : a+b+c
       * track 2 : d+e
       * you should do the following calls:
       * new_generic_material()
       * new_generic_material_track()
       * add_material_track(conc_mat, a, tcina, tcouta)
       * add_material_track(conc_mat, b, tcinb, tcoutb)
       * add_material_track(conc_mat, c, tcinc, tcoutc)
       * new_generic_material_track()
       * add_material_track(conc_mat, d, tcind, tcoutd)
       * add_material_track(conc_mat, e, tcine, tcoute)
       *
       * Note that the concrete tracks don't need to come from the same concrete material.
       */
      virtual bool CALLING_CONVENTION new_generic_material_track()=0;

      /*
       * Set a concrete track from a concrete material to be wrapped in the mxf_file.
       * The entirety of the material will be wrapped but only the media comprised between
       * [tc_in, tc_out] (both inclusive) on the given track will be played.
       * If tc_in is NULL then the media playout will be from the beginning of the concrete track.
       * If tc_out is NULL then the media playout will be until the end of the concrete track.
       * Output material will also be automatically created/updated upon calling of this function.
       * Therefore there's no guarantee that the I_generic_material retrieved from this file before
       * calling add_material will remain valid after.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION add_material_track(I_concrete_material* /* in */, 
                                                         I_concrete_track* /* in */, 
                                                         I_timecode* tc_in /* in free */, 
                                                         I_timecode* tc_out /* in free */)=0;

      /*
       * Set a concrete track from a concrete material to be wrapped in the mxf_file.
       * The entirety of the material will be wrapped but only the media comprised between
       * [tc_in, tc_out] (both inclusive) on the given track will be played.
       * If tc_in is NULL then the media playout will be from the beginning of the concrete material.
       * If tc_out is NULL then the media playout will be until the end of the concrete material.
       * Output material will also be automatically created/updated upon calling of this function.
       * Therefore there's no guarantee that the I_generic_material retrieved from this file before
       * calling add_material will remain valid after.
       * The fifth parameter is used to set the locator string (usually the path
       * to the external file) that will be included in the header metadata.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION add_external_material_track(I_concrete_material* /* in */, 
                                                                  I_concrete_track* /* in */, 
                                                                  I_timecode* tc_in /* in free */, 
                                                                  I_timecode* tc_out /* in free */, 
                                                                  const wchar_t* /* in free */)=0;

      /*
       * Call this function to state whether special processing is required at the join of two 
       * source clips of a track of the genric material. This value is used by the MXF decoder
       * to know if it can play continuously the essences. 
       * For instance if the file will contain a DV and MPEG in the same track, you should
       * set this value to false (default). Whenever special processing is required to
       * play at least one of the generic material's track, this value should be set to false.
       */
      virtual bool CALLING_CONVENTION set_continuous_decoding(bool)=0;
    
      /*
       * Return the generic material created after calling set_material_track or set_external_material_track
       * Return NULL if the generic material was not created yet.
       * Do not call add_material or add_external_material after calling this function
       * or all the changes you made to the generic material will be lost.
       */
      virtual I_generic_material* CALLING_CONVENTION get_current_generic_material()=0;
    };

    /*
     * Retrieves an I_op3c_file interface.
     * I_op3c_file** pointer to the newly allocated object
     * const char* path to the mxf file to be written.
     * Call to flush() will effectively cause the on-disk writing of
     * the file.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewOp3cFileInterface(I_op3c_file** /* out free */, 
                                                                       const char* /* in free */);

    /*
     * Retrieves an I_op3c_file interface. The flush will be performed in streaming mode.
     * I_op3c_file** pointer to the newly allocated object
     * const char* path to the mxf file to be written.
     * Call to flush() will effectively cause the on-disk writing of
     * the file.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewOp3cStreamInterface(I_op3c_file** /* out free */, 
                                                                         I_output_mxf_stream_task* /* in free */);

    /*
     * Frees an op3c_file interface.
     * I_op3c_file** pointer to the object to be freed.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeOp3cFileInterface(I_op3c_file** /* in out */);
  }
}

#endif
