//------------------------------------------------------------------------------
//  Author:   MXF Team
//  Comment:	MXF_TK: This class defines the virtual classes to be implemented
//		in order to stream MXF files.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_STREAMING_HPP
#define OC_STREAMING_HPP

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_essence_type.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    class I_mxf_file;
    class I_source_clip;
    class I_generic_material;
    class I_timecode;
    class I_mxf_file_descriptor;

    // **************************************************************************************************

    /*
     * Class I_output_mxf_stream_task
     * The API user should derive this class in order to provide its own
     * implementation of streaming while creating an mxf file.
     */
    class I_output_mxf_stream_task
    {
    public:
      /**
       * This function will be called whenever the API filled the output buffer
       * and will therefore notify the API user that some data is ready to be read
       * The first parameter will be the buffer from where the data is to be read.
       * and the second one is the buffer's size. Do not deallocate this buffer!
       * MxfTk assumes that all the data was read from the buffer once exiting this function.
       * You may also use mutexes to stop this function as long as your streaming device is not
       * ready to receive data.
       * In order to notify the end of the stream the API calls this function with a NULL buffer.
       */
      virtual void CALLING_CONVENTION data_to_read(const uint8_t* /* in */, 
                                                   size_t /* in */) =0;

      /**
       * This function can be overloaded to authorize MXFTk to seek on stream.
       * Reimplement this function to return true if you are creating an OpZero
       * file.
       */
      virtual bool CALLING_CONVENTION seekable() { return false; }

      /**
       * This function will be called whenever the API requires to move to a new location
       * on the mxf stream. It is perfectly valid not to implement this function,
       * however seekable() should always return false in that case.
       * The parameter gives the absolute seek position from the beginning of the stream.
       * The function should return the new location (should be equal to the input parameter)
       * Reimplement this function if you are creating an OpZero file.
       */
      virtual uint64_t CALLING_CONVENTION seekpos_request(uint64_t /* in */) { return 0; };
    };

    // **************************************************************************************************

    /* Class I_input_mxf_stream_task
     * The API user should derive this class in order to provide its own
     * implementation of streaming while reading an mxf file.
     */
    class I_input_mxf_stream_task
    {
    public:
      /** 
       * This function should return the size (in bytes) of the buffer that MxfTk 
       * will use to handle read and write operations on the stream. 
       * This function can be overriden. Changing the buffer size can significantly
       * improve the perfomances.
       */
      virtual size_t CALLING_CONVENTION get_buffer_size()=0;

      /**
       * This function will be called whenever the API requires data from the mxf stream
       * and will therefore notify the API user that some data is awaited.
       * Calls to I_mxf_file::write and I_mxf_file::end_of_stream should be performed to 
       * respectively feed and notify the end of the stream.
       */
      virtual void CALLING_CONVENTION data_request(I_mxf_file* /* in */, 
                                                   size_t /* in */) =0;

      /**
       * This function will be called whenever the API requires to move to a new location
       * on the mxf stream. It is perfectly valid not to implement this function,
       * however seekable should always return false in that case.
       * The parameter gives the absolute seek position from the beginning of the stream.
       * The function should return the new location (should be equal to the input parameter)
       */
      virtual uint64_t CALLING_CONVENTION seekpos_request(uint64_t /* in */) { return 0; };

      /**
       * This function notifies that some data is ready to be read on the
       * I_source_clip from the given track in the given I_generic_material.
       * The uint64_t parameter indicates the number of bytes to be read (size of the edit unit to be read).
       */
      virtual void CALLING_CONVENTION data_to_read(I_generic_material* /* in */, 
                                                   I_track* /* in */, 
                                                   I_source_clip* /* in */, 
                                                   uint64_t /* in */) =0;

      /*
       * This function gives the opportunity to continue the playout from a user-defined timecode
       * for the specified generic_material. Be aware that your streaming device should have the capability
       * to rewind if you choose to use this function (seekpos_request must be implemented!). 
       * seek_timecode must return NULL if no seek is to be performed.
       * This function will be called regularly to give several opportunities for the user to relocate
       * the file playout.
       * IMPORTANT :
       * If you try to reach a timecode that was not read yet, the reading process will simply seek
       * in the mxf stream to the closest timecode already read and will continue the reading process...
       * Finally the boolean will be set to true when MxfTk is about to exit the streaming mode.
       * You may set it to "true" to force the end of the streaming task.
       */
      virtual I_timecode* CALLING_CONVENTION seek_timecode(I_generic_material* /* in */, 
                                                           bool& /* in out */) { return NULL; };

      /**
       * The structure of an mxf file is likely to be updated as a file is played. This is the 
       * reason why several instances of an header metadata can be found in an mxf_file. 
       * Each new header metadata was written to take into account the new information 
       * the encoder received while writing the file. Therefore, the truthful header 
       * metadata is usually at the end of an mxf file. However, when streaming, this one 
       * is not reachable before the end so the API returns an updated version of the header 
       * metadata whenever a new one is detected on the stream. This helps the API user
       * starting the "decoding". However, the user should be aware that there's no guarantee 
       * that the I_mxf_file structure won't change until the stream is closed...
       * The API will call this function whenever the header metadata is updated. All the previous
       * versions of I_mxf_file* should be considered as outdated. However, to help the user
       * two boolean values indicate if the file structre is still likely to change.
       * If closed is true then the overall structure of the mxf file won't change anymore
       * (i.e no more tracks or materials will be created). Finally if the file is closed and 
       * complete this means there won't be anymore updates. The I_mxf_file just received will 
       * be the final one. If the file is closed and incomplete then the structure won't
       * change but some values might still change (track duration or track descriptor's for instance)
       * Finally the uint64_t parmaeter gives the number of bytes read from the mxf stream so far. 
       * The function should return true if you wish to stop the decoding process. This is particularly 
       * useful if you are not interested in the essence of the MXF file but simply want to stop
       * as soon as you get a valid header metadata.
       */
      virtual bool CALLING_CONVENTION updated_mxf_file(I_mxf_file* /* in */, 
                                                       bool /*closed*/ /* in */, 
                                                       bool /*complete*/ /* in */, 
                                                       uint64_t /*mxf_stream_pos*/ /* in */) 
	{
	  return false;
	}

      /**
       * Function call to set the offset of the last partition opened in reading mode. 
       */
      virtual void CALLING_CONVENTION last_partition_offset(uint64_t /* in */) {}

      /**
       * Implement this function when you are performing metadata update on the MXF stream.
       * This function will be called by MXFTk when you call the function mxf_file::flush()
       * after opening the stream for reading.
       * It will provide you the buffers (and their size) to be wriiten on the stream in order 
       * to update it (timecode redefinition - descriptive metadata).
       * The buffer must be written at the current location of the stream ( the one sent by the 
       * previous call to seekpos_request ). Note that the data may need to be appended at the end
       * of the stream.
       */
      virtual void CALLING_CONVENTION updated_data(const uint8_t* /*data*/ /* in */, 
                                                   size_t /*data_size*/ /* in */) {}

      /**
       * This function can be overloaded to set the size of the stream 
       * The function should return 0 if it cannot be determined.
       * If you set the function seekable to return true then this function
       * should also return the correct size of the stream.
       */
      virtual uint64_t CALLING_CONVENTION stream_size() { return 0; }

      /**
       * This function can be overloaded to authorize MXFTk to seek on stream.
       */
      virtual bool CALLING_CONVENTION seekable() { return false; }
    };

    /* Class I_input_partial_mxf_stream_task
     * The API user should derive this class in order to provide its own
     * implementation of streaming while reading an mxf file that will be partially
     * restored (NewPartialRestoreMxfStreamInterface)
     */
    class I_input_partial_mxf_stream_task
    {
    public:
      /** 
       * This function should return the size (in bytes) of the buffer that MxfTk 
       * will use to handle read operations on the stream. 
       * This function can be overriden. Changing the buffer size can significantly
       * improve the perfomances.
       */
      virtual size_t CALLING_CONVENTION get_buffer_size()=0;

      /**
       * This function will be called whenever the API requires data from the mxf stream
       * and will therefore notify the API user that some data is awaited.
       * Calls to I_mxf_file::write and I_mxf_file::end_of_stream should be performed to 
       * respectively feed and notify the end of the stream.
       */
      virtual void CALLING_CONVENTION data_request(I_mxf_file* /* in */, 
                                                   size_t /* in */) =0;

      /**
       * This function will be called whenever the API requires to move to a new location
       * on the mxf stream. It is perfectly valid not to implement this function,
       * however seek_timecode should always return NULL in that case.
       * The parameter gives the absolute seek position from the beginning of the stream.
       * The function should return the new location (should be equal to the input parameter)
       */
      virtual uint64_t CALLING_CONVENTION seekpos_request(uint64_t /* in */) { return 0; };

      /**
       * This function can be overloaded to set the size of the stream 
       * The function should return 0 if this cannot be determined.
       */
      virtual uint64_t CALLING_CONVENTION stream_size() { return 0; }

      /**
       * This function can be overloaded to authorize MXFTk to seek on stream.
       */
      virtual bool CALLING_CONVENTION seekable() { return false; }
    };

    // **************************************************************************************************

    /* Class I_essence_stream_task
     * The API user should derive this class in order to provide its own
     * implementation of streaming while creating a concrete material with 
     * NewStreamMaterialInterface.
     */
    class I_essence_stream_task
    {
    public:
      I_essence_stream_task():_conc_mat(NULL) {}

      /**
       * This function should return true if you choose to perform the streaming in passive mode
       * or false if you choose active mode:
       * PASSIVE MODE :
       * MXFTk notifies you when it requires some data (thanks to the function data_request). You 
       * must provide an implementation of the function data_request in that case and perform write operations 
       * only within this function.
       * ACTIVE MODE :
       * You can feed MXFTk with data whenever you want using the function write(). The function data_request will never
       * be called while in active mode.
       */
      virtual bool CALLING_CONVENTION get_passive_mode() =0;

      /**
       * This function should return true if you choose to perform the streaming in blocking active mode
       * or false if you choose buffering active mode:
       * BLOCKING MODE :
       * Available in active mode only. MXFTk perform no copies, user has to feed each essence, 
       * one full frame after the other, etc. in the expected order
       * BUFFERING MODE :
       * Data is stored in an internal buffer that allows user to feed essence 
       * without respecting frame size and order.
       */
      virtual bool CALLING_CONVENTION get_active_blocking_mode(unsigned int /* in */) { return false; }

      /** 
       * This function should return the total number of simultaneous threads to be embedded in
       * a single material (just as several files may be required to build a concrete_material,
       * several streams may be required.
       */
      virtual unsigned int CALLING_CONVENTION get_total_number_of_streams() =0;

      /**
       * This function may be optionally derived to return the essence_type and edit rate
       * of the corresponding stream. THIS FUNCTION MUST BE RE-IMPLEMENTED in your derived object. 
       * You should set the essence source corresponding to the stream id.
       * If the essence source is an MPEG Transport Stream or MPEG Program Stream you should
       * also indicate how many elementary streams it includes. This value is ignored otherwise.
       * The boolean is_video indicates if the source is a video or audio stream.
       * In the case of DV essence source, is_video should be read as "is_video_only", set 
       * it to true only if there's no audio in the DV or if you don't want to wrap the audio.
       * Furthermore D10 and MPEG ES streams need to precise if they embed video or audio.
       */
      virtual void CALLING_CONVENTION get_stream_info(unsigned int /*stream_id*/ /* in */, 
						      essence_source& s /* out */, 
						      bool& is_video /* out */, 
						      unsigned int& nb_mpeg_es /* out */) 
	{ 
	  s = ess_unknown; // unknown essence -> will ignore the call to this function 
	  is_video = true;
	  nb_mpeg_es = 0;
	}

      /**
       * This function must be implemented if you implemented get_stream_info and some of the
       * streams include an MPEG Transport Stream or MPEG Program Stream.
       */
      virtual void CALLING_CONVENTION get_mpeg_es_info(unsigned int /*stream_id*/ /* in */, 
						       unsigned int /*sub_mpeg_es_id*/ /* in */, 
                                                       essence_source& s, /* ess_mpeg_es or ess_aes3 only*/ /* out */
						       bool& is_video) /* if ess_mpeg_es states whether it's audio or video ES */ /* out */
	{
          s = ess_mpeg_es;
	  is_video = true;
	}

      /**
       * Functions used to set informations which are not defined is the essence stream
       * ex : uncompressed raw, PCM without wave header ...
       */
      virtual void CALLING_CONVENTION get_video_mxf_file_descriptor(unsigned int /*stream_id*/ /* in */, 
								    I_mxf_file_descriptor *& descriptor /* out */)
	{
	  descriptor = NULL;
	}

      /**
       * Functions used to set informations which are not defined is the essence stream
       * ex : uncompressed raw, PCM without wave header ...
       */
      virtual void CALLING_CONVENTION get_audio_mxf_file_descriptor(unsigned int /*stream_id*/ /* in */, 
								    I_mxf_file_descriptor *& descriptor /* out */)
	{
	  descriptor = NULL;
	}

      /** 
       * This function should return the size (in bytes) of the buffer that MxfTk 
       * will use to handle read and write operations on the stream. 
       * This function must be overriden. For performance reason,
       * it is better to set a buffer size at least as big as the biggest buffer
       * you will provideduring the streaming process.
       */
      virtual size_t CALLING_CONVENTION get_buffer_size(unsigned int /* in */)=0;

      /**
       * THIS FUNCTION WILL BE CALLED ONLY IF PASSIVE MODE IS SELECTED.
       * This function will be called whenever the API request data from the stream
       * The first parameter will be the id of the stream (comprised between 0 and 
       * the value returned by get_total_number_of_streams().
       * The second parameter is a pointer to the buffer where you should write data from the corresponding stream.
       * The third parameter indicates the maximum number of bytes you can write in the buffer.
       * The return value should be the number of bytes effectively written in the buffer.
       * If the return value is set to 0 then MXFTk will assume that it reached the end of the stream.
       * You may also use mutexes to stop this function as long as data is not available from
       * the streaming device.
       */
      virtual size_t CALLING_CONVENTION data_request(unsigned int /* in */, 
                                                     uint8_t* /* in out */, 
                                                     size_t /* in */) =0;

      /**
       * This function can be overloaded to set the size of the stream 
       * identified by the unsigned int id
       * The function should return 0 if it cannot be determined.
       */
      virtual uint64_t CALLING_CONVENTION stream_size(unsigned int /* in */) { return 0; }

      /**
       * This function can be overloaded to set the duration of the stream 
       * identified by the unsigned int id
       * The function should return -1 if it cannot be determined.
       */
      virtual int64_t CALLING_CONVENTION duration(unsigned int /* in */) { return -1; }

      /**
       * Returns the associated I_concrete_material* - do not reimplement
       */
      I_concrete_material* CALLING_CONVENTION get_concrete_material() { return _conc_mat; } 

      /**
       * Sets the concrete material / called by MXFTk - do not reimplement!
       */
      void CALLING_CONVENTION set_concrete_material(I_concrete_material* conc_mat) { _conc_mat = conc_mat; } 

    private:

      // Pointer to the concrete material created by the instance
      // of this class.
      I_concrete_material* _conc_mat;
    };

    // **************************************************************************************************
    /* Class I_crypted_essence_stream_task
     * The API user should derive this class in order to provide its own
     * implementation of streaming while creating a crypted concrete material with 
     * NewStreamMaterialInterface.
     */
    class I_crypted_essence_stream_task : public I_essence_stream_task
    {
    public:
      I_crypted_essence_stream_task():I_essence_stream_task() {}

      /**
       * Returns true if stream_id is crypted. If not, following functions won't be called 
       * for this stream id
       */
      virtual bool CALLING_CONVENTION do_crypt(unsigned int /*stream_id*/ /* in */) = 0;

      /**
       * Returns the plaintext offset
       * plainttext offset is the offset of the crypted data within a frame.
       * it is the size of the plaintext at start of each frame.
       */
      virtual uint32_t CALLING_CONVENTION get_plaintext_offset(unsigned int /*stream_id*/ /* in */) { return 0; }

      /**
       * Returns the cryptographic key id, this value is stored in mxf file
       * and is used by mxf decoder to retrieve cipher key.
       * an external key manager give the cipher key corresponding to a cryptographic key id
       */
      virtual const uint8_t * CALLING_CONVENTION get_cryptographic_key_id(unsigned int /*stream_id*/ /* in */) = 0;

      /**
       * Returns the cipher key.
       * Cipher key is used to crypt data but is not stored in mxf file.
       * Decoding is possible by giving the same cipher to the decoder.
       * Cipher key is retrieved from the cryptographic key id and is provided by the key manager.
       */
      virtual const uint8_t * CALLING_CONVENTION get_cipher_key(unsigned int /*stream_id*/ /* in */) = 0;
    };


    // **************************************************************************************************
    /* Class I_input_metadata_stream_task
     * The API user should derive this class in order to provide its own
     * implementation of metadata streaming while creating an mxf file.
     */
    class I_input_metadata_stream_task
    {
    public:
      /**
       * This function lets the user update the content of the mxf file while creating it 
       * during a streaming process. Therefore it becomes possible to add descriptive metadata
       * on-the_fly as the file is created. This function will be called whenever a partition has been 
       * written. Therefore, the user must understand that the key parameter to use efficiently
       * this function is "PREFERRED_PARTITION_DURATION". The larger the partition size, the less often
       * the function update will be called; on the contrary, the smaller the partition size, the 
       * more often the user will have the opportunity to update the file.
       * The parameter is a pointer on the metadata track to be updated and the integer value is the 
       * current position (edit unit) on this track.
       */
      virtual void CALLING_CONVENTION update(I_generic_material* /* in */, 
                                             I_track* /* in */, 
                                             int64_t /* in */) =0;
    };

  }
}

#endif
