//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: An example to read the content of an mxf stream.
//  This example simulates a possible implementation for reading data
//  from a linear storage device.
//  You should be familiar with the non-streaming interfaces
//  before going through this example. This is an advanced example(!)
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/I_mxf_error.hpp>
#include <mxf_tk/I_mxf_file.hpp>

#include <iostream>
#include <sstream>
#include <map>

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

bool closed_and_complete = false;

// This is the size of the buffer where the mxf stream is read 
// (buffer from the mxf linear-storage device).
#define BUF_SIZE 16384

// We derive the pure virtual class I_input_mxf_stream_task.
// This class provides several functions that will be automatically 
// called by MxfTk whenever some data is ready to be read or
// whenever some data is required from the mxf stream.
// For the conveniency of the demonstration, this implementation reads
// mxf data from a file however the same mechanism could
// be applied to a true streaming device.
class I_user_input_mxf_stream_task : public I_input_mxf_stream_task
{
public:
  // Constructor -> mxf file (mxf streaming device buffer)
  I_user_input_mxf_stream_task(std::string);

  // Destructor -> close file (free streaming device buffer)
  virtual ~I_user_input_mxf_stream_task();

  // Sets the size of MxfTk internal buffer used to bufferize the
  // mxf stream. MxfTk user is free to change its value.
  // Increasing the buffer size may help to speed up the process
  // Here, it is set to the same size as the streaming device buffer (BUF_SIZE). 
  virtual size_t CALLING_CONVENTION get_buffer_size() { return BUF_SIZE; };

  // MxfTk will call this function whenever it requires soma data from the
  // mxf stream. User should provide an implementation "feeding" Mxftk
  // with the content of the storage device's buffer.
  virtual void CALLING_CONVENTION data_request(I_mxf_file*, size_t);

  // MxfTk will call this function when a seek operation is required on 
  // the input mxf stream. This call will occur only if a seek operation
  // was requested by the API user (thanks to seek_timecode). Therefore
  // it is perfectly valid to have a nil implementation for this function 
  // if you do not wish to perform seek or if your storage device is not
  // capable of seeking.
  virtual uint64_t CALLING_CONVENTION seekpos_request(uint64_t);

  // MxfTk will call this function whenever some essence data must be read.
  // It corresponds to the output of the mxf file.
  virtual void CALLING_CONVENTION data_to_read(I_generic_material*, I_track*, I_source_clip*, uint64_t);

  // You should provide an implementation if you wish to continue mxf
  // decoding from a given timecode. Seek operations are only allowed to
  // positions of the file that were already read. If you ask to seek to
  // a timecode that was not read yet, this function will inoperate.
  virtual I_timecode* CALLING_CONVENTION seek_timecode(I_generic_material*, bool&);

  // This function is called whenever a new header metadata is met in the file.
  // Previous mxf_files become obsolete.
  virtual bool CALLING_CONVENTION updated_mxf_file(I_mxf_file*, bool closed, bool complete, uint64_t);
protected:
  // storage device buffer
  uint8_t* buffer;
  // file handle (storage device)
  FILE* file;
  // filename 
  std::string _mxf_file_name;

  // A map containing the names of the output files
  // There will be one file for each output track of
  // the generic material. You may replace the output 
  // files by any other storage mechanism.
  std::map <uint32_t, FILE*> out_files;
};

// Constructor : Open the mxf storage device for reading (file) 
// and retrieve its buffer.
I_user_input_mxf_stream_task::I_user_input_mxf_stream_task(std::string mxf_file_name)
: I_input_mxf_stream_task()
{
  file = fopen(mxf_file_name.c_str(), "rb");  // open storage device
  if (!file || ferror(file))
    wcout << L"An error occured while opening the mxf file" << endl;

  buffer = new uint8_t[BUF_SIZE]; // retrieve storage device's buffer
  _mxf_file_name = mxf_file_name; // store for later use.
}

// Destructor : Frees storage device's buffer and close
// output files.
I_user_input_mxf_stream_task::~I_user_input_mxf_stream_task()
{
  delete [] buffer;
  fclose(file);

  for (map<uint32_t, FILE*>::iterator it = out_files.begin(); it != out_files.end(); it++)
  {
    FILE* f= it->second;
    fclose(f);
  }
}

// MxfTk notifies its need for data from the input mxf stream.
// Called with an interface on the mxf file being processed and
// ask for a given number of bytes.
void CALLING_CONVENTION I_user_input_mxf_stream_task::data_request(I_mxf_file* pi_mxf_file, size_t s_req)
{
  try
  {
    // Check MxfTk doesn't ask more than the storage device's buffer.
    if (s_req>BUF_SIZE)
      s_req = BUF_SIZE;

    // It is perfectly valid to give less data than the amount specified (s_req)
    // data_request() is called whenever MxfTk internal buffer is empty (all
    // the previous mxf data has been processed)

    // read data from the storage device
    // If no data is available, it is possible to use mutex to
    // stop the current process as long as no data is available from
    // the storage device.
    size_t s_read = fread(buffer, sizeof(uint8_t), s_req, file);
    if (ferror(file)) throw 1;

    // There's some data in the buffer.
    if (s_read>size_t(0))
    {
      // Then give it to MxfTk
      size_t written = pi_mxf_file->write(buffer, s_read);
      wcout << L"Mxf Stream: " << (long)s_read << " bytes written" << endl;
      if (written<s_read) throw 1; // couldn't write all  
    }
    // No more data to be processed. The file is closed then notifies it to MxfTk
    else
    {
      pi_mxf_file->end_of_stream();
      wcout << L"Mxf Stream: closed" << endl; 
    }
    throw_on_error();
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }
}

// The user asked for seeking a given timecode
// Notifies the storgae device it has to seek to a given position
// If seek timecode always returns NULL you do not need to
// provide an implementation for this function.
uint64_t CALLING_CONVENTION I_user_input_mxf_stream_task::seekpos_request(uint64_t pos)
{
  /* Possible implementation
  if (fseek(file, long(pos), SEEK_SET)!=0)
  std::cout << "Cannot seek there!" << std::endl;
  return pos;
  */
  wcout << L"Requesting a seek operation -> Invalid call" << endl;
  return pos;
}

// MxfTk notifies that some essence data must be read.
// Data from sc source clip contained by track from
// the given generic material.
// This function will be called when an edit unit (a frame)
// from the source clip can be read.
void CALLING_CONVENTION I_user_input_mxf_stream_task::data_to_read(I_generic_material*, I_track* track, I_source_clip* sc, uint64_t to_be_read)
{
  // I_generic_material's UMID and track id can be used
  // to uniquely identify the output track.

  try
  {
    // data from which track ?
    uint32_t tr_id = track->get_track_id();

    FILE* f = NULL;
    if (out_files.find(tr_id) != out_files.end())
      // Retrieve out put file for this track.
    {
      f = out_files[tr_id];
    }
    // if the output file for this track does not exist yet -> create it!
    else
    {
      stringstream ss;
      ss << (long)tr_id;
      std::string thisfile = _mxf_file_name + std::string(".") + ss.str(); 
      f = fopen(thisfile.c_str(), "wb");
      if (!f) throw 1;
      out_files[tr_id] = f;
    }

    // If you are not ready to process the output data (MwfTk decodes
    // faster than you can record then you can put asleep this
    // process with a mutex and wakes it up when ready.

    // You may also ignore to read the data and directly return.

    // temporary output buffer
    uint8_t* lbuffer = new uint8_t[65536];

    uint64_t nb_read = 0;
    while (nb_read !=to_be_read) // as long as all the data is not read
    {
      // how much data to be read?
      size_t buf_s = size_t(to_be_read - nb_read);
      // output buffer is 64Kbytes
      // cannot retrieve more at once.
      if (to_be_read - nb_read > 65536)
        buf_s = 65536;

      // Read edit unit (data) from the source clip.
      uint64_t read = sc->read_eu(lbuffer, buf_s);
      wcout << L"Track " << (long)tr_id << L" : " << (long)buf_s << L" bytes read" << endl;

      if (read==0)
      {
        break;
      }

      // Stores output data -> this can be recording device as well
      fwrite(lbuffer, 1, size_t(read), f);
      nb_read += read;
    }

    delete[] lbuffer;
    throw_on_error();
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }
}

// MxfTk wil regularly call this function to give you the opportunity to
// pursue the decoding from a given timecode (timecode data should be relative
// to the timecode material of the given generic material.
// This function should return the timecode where the decoding process should continue.
// Return NULL to continue normally.
// last chance will be set to true when reaching the end of the mxf stream. This will be the 
// last opportunity to seek past before the streaming process ends up
// It you choose to implement this function you MUST implement seekpos_request as well.
// Remember that you may only seek to a timecode that was already reached. In the 
// opposite case the reading process will continue as is.
I_timecode* CALLING_CONVENTION I_user_input_mxf_stream_task::seek_timecode(I_generic_material*, bool&/* last_chance*/)
{
  return NULL;

  // Possible implementation for a process looping endlessly
  // Asuming start timecode is 00:00:00:00 / frame rate =25fps
  // Timecode frame rate, drop frame, etc can be retrieved from
  // I_generic_material.
  // if (last_chance)
  // {
  //  I_timecode* time = NULL;
  //  NewTimecodeInterface(&time, "00:00:00:00", 25, false);
  //  return time;
  // }
}

// A new header metadata was found in the file.
// Previous version of mxf_file is obsolete.
// You will be ensured of the completeness of the 
// I_mxf_file structure only when closed and complete will be set
// to true.
bool CALLING_CONVENTION I_user_input_mxf_stream_task::updated_mxf_file(I_mxf_file* pi_mxf_file, bool closed, bool complete, uint64_t)
{
  if (!closed)
    wcout << L"***** NEW OPENED MXF FILE STRUCTURE *****" << endl;
  else if (!complete)
    wcout << L"***** NEW CLOSED MXF FILE STRUCTURE *****" << endl;
  else
    wcout << L"******* FINAL MXF FILE STRUCTURE ********" << endl;

  closed_and_complete = closed && complete;

  // output file structure in an xml file
  pi_mxf_file->output_xml((_mxf_file_name + std::string(".xml")).c_str());
  // Do whatever you want with the mxf_file here (except read/write/seek operations
  // which should only occur in the other functions from this class).

  return false; // you may stop the decoding process by returning true here.
}

// **************************************************************************
// **************************************************************************

int main(int argc, char* argv[])
{
  if (argc < 2)
  {
    wcout << L"usage : in.mxf" << endl;
    exit(0);
  }

  I_mxf_file* pi_mxf_file = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();
    // remove commentaries to set KLV by KLV access when
    // unwrapping custom wrapped MXF files (XDCAM proxy for
    // instance).
    //ExtractCustomByKLV(true);

    wcout << L"Init mxf streaming device with " << argv[1] << endl;
    I_user_input_mxf_stream_task in_mxf(argv[1]);

    // All the decoding occurs in this function
    NewMxfStreamInterface(&pi_mxf_file, &in_mxf);
    throw_on_error();
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  FreeMxfFileInterface(&pi_mxf_file);
  return 0;
}

void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
