//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: This example illustrates opatom(2b) file creation.
//            you should provide four source files:
//            Each of these source files will be wrapped in an Opatom file.
//            SourceFile SF1 and SF2 will be played together, then followed
//            by SF3 and SF4 played together as well.
//            Possible call is:
//            example1_opatom out video1.m2v sound1.bwf video2.m2v sound2.bwf
//            For best behavior SF1 and SF2 should have the same duration;
//            SF3 and SF4 as well.
//
//            A short explanation on OpAtom files: An Opatom file embeds a single track
//            of essence. However it is possible to build complex editing of several 
//            opatom files. The editing information (generic material) becomes an
//            assembly of several OpAtom files. The files are uniquely identified
//            by their UMID. Therefore in order to get the complete output all the 
//            OpAtom files that were generated together must be provided.
//            
// 
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/I_concrete_material.hpp>
#include <mxf_tk/I_opatom_file.hpp>

#include <iostream>
#include <string>

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

int main(int argc, char* argv[])
{
  if (argc != 3 )
  {
    wcout << L"usage : out_basename sourcefile1" << endl;
    exit(0);
  }

  // The four OpAtom files and their associated material
  I_opatom_file* pi_atom_file1 = NULL;
  I_opatom_file* pi_atom_file2 = NULL;
  I_opatom_file* pi_atom_file3 = NULL;
  I_opatom_file* pi_atom_file4 = NULL;
  I_concrete_material* pi_conc_mat1 = NULL;
  I_concrete_material* pi_conc_mat2 = NULL;
  I_concrete_material* pi_conc_mat3 = NULL;
  I_concrete_material* pi_conc_mat4 = NULL;

  // The object used to link and edit OpAtom files.
  I_opatom_assembler* pi_assembler = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    wcout << L"Create the opatom mxf files" << endl;
    NewOpAtomFileInterface(&pi_atom_file1, (std::string(argv[1])+std::string("_1.mxf")).c_str());
    NewOpAtomFileInterface(&pi_atom_file2, (std::string(argv[1])+std::string("_2.mxf")).c_str());
    NewOpAtomFileInterface(&pi_atom_file3, (std::string(argv[1])+std::string("_3.mxf")).c_str());
    NewOpAtomFileInterface(&pi_atom_file4, (std::string(argv[1])+std::string("_4.mxf")).c_str());
    throw_on_error();

    pi_atom_file1->set_parameter(KAG_SIZE, "256");

    wcout << L"Create source material 1" << endl;
    locators* pi_loc = NULL;
    GetEmptyLocatorInterface(&pi_loc);
    pi_loc->add((const char**)&(argv[2]));
    NewConcreteMaterialInterface(&pi_conc_mat1, pi_loc, clip_wrapping);
    FreeLocatorInterface(&pi_loc);
    throw_on_error();

    wcout << L"Attach material to the opatom file 1" << endl;
    pi_atom_file1->set_material(pi_conc_mat1);

    cout << L"Create source material 2" << endl;
    pi_loc = NULL;
    GetEmptyLocatorInterface(&pi_loc);
    pi_loc->add((const char**)&(argv[3]));
    NewConcreteMaterialInterface(&pi_conc_mat2, pi_loc, frame_wrapping);
    FreeLocatorInterface(&pi_loc);
    throw_on_error();

    wcout << L"Attach material to the opatom file 2" << endl;
    pi_atom_file2->set_material(pi_conc_mat2); 

    wcout << L"Create source material 3" << endl;
    pi_loc = NULL;
    GetEmptyLocatorInterface(&pi_loc);
    pi_loc->add((const char**)&(argv[4]));
    NewConcreteMaterialInterface(&pi_conc_mat3, pi_loc, frame_wrapping);
    FreeLocatorInterface(&pi_loc);
    throw_on_error();

    wcout << L"Attach material to the opatom file 3" << endl;
    pi_atom_file3->set_material(pi_conc_mat3); 

    wcout << L"Create source material 4" << endl;
    pi_loc = NULL;
    GetEmptyLocatorInterface(&pi_loc);
    pi_loc->add((const char**)&(argv[5]));
    NewConcreteMaterialInterface(&pi_conc_mat4, pi_loc, frame_wrapping);
    FreeLocatorInterface(&pi_loc);
    throw_on_error();

    wcout << L"Attach material to the opatom file 4" << endl;
    pi_atom_file4->set_material(pi_conc_mat4);

    // Build Opatom editing 
    // First link the files that should be played together;
    // then serializes file1
    wcout << L"Create OpAtom assembler" << endl;
    NewOpAtomAssemblerInterface(&pi_assembler);
    wcout << L"Parallelize (file1 // file2)" << endl;
    pi_assembler->parallelize(pi_atom_file1, pi_atom_file2);
    wcout << L"Parallelize (file3 // file4)" << endl;
    pi_assembler->parallelize(pi_atom_file3, pi_atom_file4);
    throw_on_error();

    wcout << L"Serialize (file1 // file2) + (file3 // file4)" << endl;
    pi_assembler->serialize(pi_atom_file1, pi_atom_file3);
    throw_on_error();
    // because we performed (file1 // file2) + (file3 // file4)
    // serializing file1 with file3 automatically serializes file2 with file4
    // Moreover pi_assembler->serialize(pi_atom_file1, pi_atom_file3);  or
    // pi_assembler->serialize(pi_atom_file2, pi_atom_file3); or
    // pi_assembler->serialize(pi_atom_file1, pi_atom_file4); or
    // pi_assembler->serialize(pi_atom_file2, pi_atom_file4);
    // would produce the same editing.

    wcout << L"Write files on disk" << endl;
    pi_atom_file1->flush(); // after serializing/parallelizing files the call to flush 
                            // on one I_opatom_file will flush ALL the I_opatom_files.
    throw_on_error();

    wcout << L"Wait termination" << endl;
    pi_atom_file1->wait_end_flush_thread();
    throw_on_error();
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  // the files have been assembled so deleting one of them 
  // will destroy all of them.
  FreeOpAtomFileInterface(&pi_atom_file1);

  return 0;
}

void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
