//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK : Init. function.
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_MXF_TK_HPP
#define OC_MXF_TK_HPP

#ifndef NULL
# define NULL 0
#endif

#if defined(WIN32)
# define CALLING_CONVENTION __stdcall
#else
# define CALLING_CONVENTION
#endif

#if defined(WIN32)
# ifdef MXF_TK_STATIC
#  define mxf_tk_API
# else
#  ifdef MXF_TK_EXPORTS
#   define mxf_tk_API __declspec(dllexport)
#  else
#   define mxf_tk_API __declspec(dllimport)
#  endif
# endif
#else
# if defined(sun) || defined(__sun__)
#  define mxf_tk_API
# else
#  define mxf_tk_API __attribute__ ((visibility("default")))
# endif
#endif

#include <stddef.h>

#include <mxf_tk/mxftk_defs.hpp>
#include <mxf_tk/redef.hpp>
#include <mxf_tk/inttypes.hpp>
#include <mxf_tk/I_mxf_error.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /*
     * Intializes MXF_TK before any process.
     * This function should ALWAYS be the first one to be called
     * before using MXF_TK interfaces. Failure to do so will possibly
     * result in an hectic behavior.
     * This function will fail if you didn't set the environment variable MXF_HOME
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION InitMXF_TK(void);
	
    /*
     * Intializes MXF_TK before any process.
     * This function should ALWAYS be the first one to be called
     * before using MXF_TK interfaces. Failure to do so will possibly
     * result in an hectic behavior.
     * You don't need to set the environment variable MXF_HOME when calling
     * this function, however you should provide the complete path toward the directory 
     * containing your MXFTk license file.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION InitMXF_TK2(const char* /* in free */);

    /**
     * Set a new metadata dictionary.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION SetDictionary(const char* /* in free */);

     /**
     * Set internal buffers size.
     * Call this function to override the default settings and reduce/increase memory usage.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION SetBuffersSize(uint32_t /* in */);

    /**
     * By setting the boolean parameter to true, the AES3 (mono or stereo) sources embedded
     * in an MXF file will be extracted as WAVE.
     */
    extern "C" mxf_tk_API void CALLING_CONVENTION ExtractAesToWave(bool /* in */);

    /**
     * By setting the boolean parameter to true, the 8-channel AES (D10) sources embedded
     * in an MXF file will be extracted as 8-channel WAVE.
     */
    extern "C" mxf_tk_API void CALLING_CONVENTION ExtractAes8ToWave(bool /* in */);

    /**
     * By setting the boolean parameter to true, the WAVE (mono or stereo) sources embedded
     * in an MXF file will be wrapped as AES audio data.
     */
    extern "C" mxf_tk_API void CALLING_CONVENTION WrapWaveAsAes(bool /* in */);

    /**
     * By setting the boolean parameter to true, the WAVE (mono or stereo) sources embedded
     * in an MXF file will be wrapped as 8-channel AES audio data.
     */
    extern "C" mxf_tk_API void CALLING_CONVENTION WrapWaveAsAes8(bool /* in */);

    /**
     * By setting the boolean parameter to true, the WAVE (mono or stereo) sources embedded
     * in an MXF file will be extracted as RF64 WAVE files (default is false).
     */
    extern "C" mxf_tk_API void CALLING_CONVENTION ForceRF64(bool /* in */);

    /**
     * Use this function to set the essence extraction bahavior when 
     * performing MXF unwrapping of custom wrapped MXF files (i.e.
     * XDCam proxy)
     * boolean set to false (default) -> extraction is performed 
     * edit by edit unit.
     * boolean set to true -> extraction is performed KLV by KLV and hence
     * each KLV will contain several edit units. This mode will improve 
     * performances!
     */
    extern "C" mxf_tk_API void CALLING_CONVENTION ExtractCustomByKLV(bool /* in */);

    /*
     * Defines a rational (numerator / denominator)
     * MXF standard does not allow usage of floating point values
     */
    class rational
    {
    public:
      virtual uint32_t CALLING_CONVENTION get_num()=0;
      virtual uint32_t CALLING_CONVENTION get_den()=0;
      virtual double CALLING_CONVENTION get_double()=0;
    };

    /*
     * Returns an interface on a rational
     * first and second uint32_t define numerator and denominator
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewRationalInterface(rational** /* out free */, 
                                                                       uint32_t /* in */, 
                                                                       uint32_t /* in */);
	
    /*
     * Frees an interface on a rational
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeRationalInterface(rational** /* in out */);

    /*
     * Defines a list of I_opatom_file
     */
    class I_mxf_file;
    class mxf_file_list
    {
    public:
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual I_mxf_file** CALLING_CONVENTION next()=0;
      virtual void CALLING_CONVENTION add(I_mxf_file** /* in free */)=0;
      virtual void CALLING_CONVENTION remove(I_mxf_file** /* in free */)=0;
    };

    /*
     * Returns an interface on an empty track_list
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION GetEmptyMxfFileListInterface(mxf_file_list** /* out free */);
	
    /*
     * Frees an interface on a track_list
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeMxfFileListInterface(mxf_file_list** /* in out */);

    /*
     * Defines a list of I_generic_material
     */
    class I_generic_material;
    class generic_list
    {
    public:
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual I_generic_material** CALLING_CONVENTION next()=0;
    };

    /*
     * Defines a list of I_concrete_material
     */
    class I_concrete_material;
    class concrete_list
    {
    public:
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual I_concrete_material** CALLING_CONVENTION next()=0;
    };

    /*
     * Defines a list of I_property
     */
    class I_property;
    class properties_list
    {
    public:
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual I_property** CALLING_CONVENTION next()=0;
    };

    /*
     * Defines a list of I_track
     */
    class I_track;
    class track_list
    {
    public:
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual I_track** CALLING_CONVENTION next()=0;
      virtual void CALLING_CONVENTION add(I_track** /* in free */)=0;
      virtual void CALLING_CONVENTION remove(I_track** /* in free */)=0;
    };

    /*
     * Returns an interface on an empty track_list
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION GetEmptyTrackListInterface(track_list** /* out free */);
	
    /*
     * Frees an interface on a track_list
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeTrackListInterface(track_list** /* in out */);

    /*
     * Defines a list of I_metadata
     */
    class I_metadata;
    class metadata_list
    {
    public:
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual I_metadata** CALLING_CONVENTION next()=0;
    };

    /*
     * Defines a list of I_mxf_error
     */
    class I_mxf_error;
    class error_list
    {
    public:
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual I_mxf_error** CALLING_CONVENTION next()=0;
    };

    /*
     * Defines a list of const char* defining essence files location
     */
    class locators
    {
    public:
      virtual void CALLING_CONVENTION add(const char**)=0;
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual const char** CALLING_CONVENTION next()=0;
    };

    /*
     * Returns an interface on an empty locators list
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION GetEmptyLocatorInterface(locators** /* out free */);
	
    /*
     * Frees an interface on a locators list
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeLocatorInterface(locators** /* in out */);


    /*
     * Defines a crypted_locator_element 
     * location is a file name
     * plaintext_offset is the offset in each essence frame where starts the encryption
     * cryptographic_key_id is an UUID 
     * cipher_key is a key used during encryption and decryption
     */
    class crypted_locator_element
    {
    public:
      virtual const char* CALLING_CONVENTION get_location() const = 0;
      virtual uint64_t CALLING_CONVENTION get_plaintext_offset() const = 0;
      // 16 bytes array (UUID)
      virtual uint8_t* CALLING_CONVENTION get_cryptographic_key_id() const = 0;
      // 16 bytes array
      virtual uint8_t* CALLING_CONVENTION get_cipher_key() const = 0;
    };

    /*
     * Returns an interface on a crypted_locator_element
     * location defines essence files location
     * cryptographic_key_id is the embeded identifier which let decoder retrieve cipher_key
     *  from a key management module (SMPTE423M)
     * cipher_key is the key used to crypt data, this key is not stored in mxf file.
     * lengths of both cryptographic_key_id and cipher_key are 16 bytes.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewCryptedLocatorElementInterface(crypted_locator_element** /* out free */, 
                                                                                    const char* /* in free */, 
                                                                                    const uint64_t /* in */, 
                                                                                    const uint8_t* /* in free */, 
                                                                                    const uint8_t* /* in free */);
	
    /*
     * Frees an interface on a crypted_locator_element
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeCryptedLocatorElementInterface(crypted_locator_element** /* in out */);


    /*
     * Defines a list of crypted_locator_element
     */
    class crypted_locators
    {
    public:
      virtual void CALLING_CONVENTION add(crypted_locator_element** /* in free */)=0;
    };

    /*
     * Returns an interface on an empty locators list
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION GetEmptyCryptedLocatorInterface(crypted_locators** /* out free */);
	
    /*
     * Frees an interface on a locators list
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeCryptedLocatorInterface(crypted_locators** /* in out */);

    /*
     * Defines a list of I_timecode_component (to be used in I_timecode_material)
     */
    class I_timecode_component;
    class ordered_timecode_list
    {
    public:
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual I_timecode_component** CALLING_CONVENTION next()=0;
    };

    /*
     * Defines a list of I_track_item
     */
    class I_track_item;
    class ordered_track_item_list
    {
    public:
      virtual int64_t CALLING_CONVENTION size()=0;
      virtual void CALLING_CONVENTION begin()=0;
      virtual I_track_item** CALLING_CONVENTION next()=0;
    };
  }
}

#endif
