//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: This example illustrates mxf file creation while streaming
//                      the output mxf file.
//		        We use files to feed the stream in this example but you 
//                      may use a a real stream (a Firewire port for instance).
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/I_concrete_material.hpp>
#include <mxf_tk/I_op1a_file.hpp>

#include <iostream>
#include <sstream>

// Size of the input stream buffer
#define BUF_SIZE 16384

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

//***********************************************************************************
// Derivation of the pure virtual class from MxfTk.
// This class should be used to retrieve the mxf stream
// as it is created by MxfTk.
// It can be used to store the data in a storage device for instance
// Here we provide an implementation with files for conveniency.
class I_user_output_mxf_stream_task : public I_output_mxf_stream_task
{
public:
  // Constructor -> Init output mxf stream 
  I_user_output_mxf_stream_task(std::string);
  // Destructor -> Close output mxf stream
  virtual ~I_user_output_mxf_stream_task();

  // This function notifies that size_t bytes are ready to be read
  // from MxfTk output buffer.
  virtual void CALLING_CONVENTION data_to_read(const uint8_t*, size_t);
protected:
  // handle to mxf storage device.
  FILE* file;
};

// Init output storage device
I_user_output_mxf_stream_task::I_user_output_mxf_stream_task(std::string mxf_file_name)
: I_output_mxf_stream_task()
{
  file = fopen(mxf_file_name.c_str(), "wb");
  if (ferror(file))
    wcout << L"An error occured while opening the mxf file " << mxf_file_name.c_str() << endl;
}

I_user_output_mxf_stream_task::~I_user_output_mxf_stream_task()
{
}

// Returns a buffer of s bytes containing the output mxf stream
void CALLING_CONVENTION I_user_output_mxf_stream_task::data_to_read(const uint8_t* buffer, size_t s)
{
  if (buffer)
    // There's a buffer
  {
    // Then writes the data to the storage device
    size_t s_written = fwrite(buffer, sizeof(uint8_t), s, file);
    if (ferror(file) || s_written!=s)
      wcout << L"Error while writing output mxf stream" << std::endl;
    wcout << L"Output MXF stream: " << (long)s_written << L" bytes read" << endl; 
  }
  else
    // No buffer -> that means the stream is closed
  {
    fclose(file);
    wcout << L"Output MXF stream: closed" << endl;
  }
}

//*************************************************************************************************************
//*************************************************************************************************************

int main(int argc, char* argv[])
{
  if (argc < 3 )
  {
    cout << "usage : out.mxf source file [...source file]" << endl;
    exit(0);
  }

  I_op1a_file* pi_op1a_file = NULL;
  bool b_delete_op1a = false;
  I_concrete_material* pi_conc_mat = NULL;
  bool b_delete_mat = false;
  // Path to the source files to be wrapped
  locators* pi_loc = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    wcout << L"Create the mxf file streaming device" << endl;
    I_user_output_mxf_stream_task mxf_task(argv[1]);

    cout << "Create the op1a mxf file" << endl;
    NewOp1aStreamInterface(&pi_op1a_file, &mxf_task);
    throw_on_error();
    b_delete_op1a = true;

    // Use this to partition and repeat header metadata throughout the file
    /*
    wcout << L"Set partition_size to 2 seconds" << endl;
    pi_op1a_file->set_parameter(PREFERRED_PARTITION_DURATION, "2.0");
    wcout << L"Set header metadata repetition" << endl;
    pi_op1a_file->set_parameter(HEADER_REPETITION, "5");
    // header metadata will be repeated every five partitions
    */

    wcout << L"Build source files list" << endl;
    GetEmptyLocatorInterface(&pi_loc);
    throw_on_error();
    for (int i=2; i<argc; i++)
    {
      pi_loc->add((const char**)&(argv[i]));
    }

    wcout << L"Create source material" << endl;
    NewConcreteMaterialInterface(&pi_conc_mat, pi_loc, frame_wrapping);
    throw_on_error();
    b_delete_mat = true;

    wcout << L"Attach material to the file" << endl;
    pi_op1a_file->set_material(pi_conc_mat); 
    b_delete_mat =false; // pi_conc_mat will be deleted upon destruction of pi_op1a_file

    wcout << L"Launch the wrapping process" << endl;
    pi_op1a_file->flush();
    throw_on_error();
    
    wcout << L"Wait for its completion" << endl;
    pi_op1a_file->wait_end_flush_thread();
    throw_on_error();

    wcout << L"Close Process" << endl;
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  if (b_delete_op1a)
    FreeOp1aFileInterface(&pi_op1a_file);
  if (b_delete_mat)
    FreeConcreteMaterialInterface(&pi_conc_mat);
  
	return 0;
}

void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}

