//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: This example illustrates op1c mxf file creation.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_concrete_material.hpp>
#include <mxf_tk/I_op1c_file.hpp>

#include <iostream>
#include <string>

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

int main(int argc, char* argv[])
{
  if (argc < 6 )
  {
    cout << "usage : out.mxf source_video_file source_audio_file source_audio_file [...source_audio_file]" << endl;
    cout << "This example will wrap all the source files in the same concrete material." << endl;
    cout << "but will produce an output material(video+audio) for each audio source." << endl;
    exit(0);
  }

  // the op1c mxf file we are going to write.
  I_op1c_file* pi_op1c_file = NULL;
  // Path to the source files to be wrapped
  locators* pi_loc = NULL;
  // the interface to manipulate the source to be wrapped.
  I_concrete_material* pi_conc_mat = NULL;
  // this boolean is set to true if pi_conc_mat needs to be destroyed
  bool delete_conc_mat = false;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    wcout << L"Create the op1a mxf file interface" << endl;
    NewOp1cFileInterface(&pi_op1c_file, argv[1]);
    throw_on_error();

    // Use this to tell MxfTk to generate index tables for this file
    // Will not operate with MxfTk Standard version
    wcout << L"Activate Index Tables" << endl;
    pi_op1c_file->set_parameter(INDEX_TABLE, "True");

    wcout << L"Build source files list" << endl;
    GetEmptyLocatorInterface(&pi_loc);
    throw_on_error();
    for (int i=2; i<argc; i++)
    {
      pi_loc->add((const char**)&(argv[i]));
    }

    wcout << L"Create source material" << endl;
    NewConcreteMaterialInterface(&pi_conc_mat, pi_loc, frame_wrapping);
    FreeLocatorInterface(&pi_loc);
    throw_on_error();
    delete_conc_mat = true;

    wcout << L"Get list of source tracks" << endl;
    I_concrete_track* picture_track = NULL;
    track_list* trl = pi_conc_mat->source_tracks();
    if (trl)
    {
      trl->begin();
      I_track** ptr = NULL;
      while ((ptr = trl->next()) != NULL)
      {
        I_concrete_track* ctr = (I_concrete_track*)*ptr;
        if (ctr->type() == picture)
        {
          picture_track = ctr;
          break;
        }
      }
      pi_conc_mat->free_tracks(trl);
    }

    if (!picture_track)
    {
      wcout << L"You need to include at least one video source to run this wrapper.";
      throw 1;
    }

    trl = pi_conc_mat->source_tracks();
    if (trl)
    {
      trl->begin();
      I_track** ptr = NULL;
      while ((ptr = trl->next()) != NULL)
      {
        I_concrete_track* ctr = (I_concrete_track*)*ptr;
        if (ctr != picture_track)
        {
          pi_op1c_file->new_generic_material();
          pi_op1c_file->add_material_track(pi_conc_mat, picture_track);
          delete_conc_mat = false; 
          pi_op1c_file->add_material_track(pi_conc_mat, ctr);
          throw_on_error();
        }
      }
      pi_conc_mat->free_tracks(trl);
    }

    wcout << L"Write mxf file on disk" << endl;
    pi_op1c_file->flush();
    throw_on_error();

    wcout << L"Wait termination" << endl;
    pi_op1c_file->wait_end_flush_thread();
    throw_on_error();
  }
  catch (int)
  {
    wcout << "An error occured!" << std::endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  wcout << L"Close Process" << endl;
  if (pi_op1c_file)
    FreeOp1cFileInterface(&pi_op1c_file);
  if (pi_conc_mat && delete_conc_mat)
    FreeConcreteMaterialInterface(&pi_conc_mat);
  if (pi_loc)
    FreeLocatorInterface(&pi_loc);

  return 0;
}

void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
