//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: An example to read the content of opatom files.
//  Illustrates opatom files synchronization. Also illustrates
//  a full traversal and console output of an mxf_file structure.
//  Extracts the content of each OpAtom file
//  A short explanation on OpAtom files: An Opatom file embeds a single track
//  of essence. However it is possible to build complex editing of several 
//  opatom files. The editing information (generic material) becomes an
//  assembly of several OpAtom files. The files are uniquely identified
//  by their UMID. Therefore in order to get the complete output all the 
// OpAtom files that were generated together must be provided.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/I_metadata.hpp>
#include <mxf_tk/I_concrete_track.hpp>
#include <mxf_tk/I_timecode_material.hpp>
#include <mxf_tk/I_metadata_material.hpp>
#include <mxf_tk/I_value.hpp>
#include <mxf_tk/I_mxf_error.hpp>
#include <mxf_tk/I_mxf_file.hpp>
#include <mxf_tk/I_op1a_file.hpp>
#include <mxf_tk/I_opatom_file.hpp>

#include <iostream>
#include <sstream>
#include <fstream>

#define APPLICATION_VERSION "http://www.smpte-ra.org/schemes/434/200X/properties/RP210:ApplicationVersion"
#define MODIFICATION_DATE "http://www.smpte-ra.org/schemes/434/200X/properties/RP210:FileModificationDate"
#define APPLICATION_SUPPLIER "http://www.smpte-ra.org/schemes/434/200X/properties/RP210:ApplicationSupplierName"
#define APPLICATION_NAME "http://www.smpte-ra.org/schemes/434/200X/properties/RP210:ApplicationName"
#define APPLICATION_VERSION_STRING "http://www.smpte-ra.org/schemes/434/200X/properties/RP210:ApplicationVersionString"

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

// Called to display the MXffile version
void read_file_version(const I_metadata*);

// Displays history of modifications on the file.
void read_history(I_mxf_file*);

// Displays the output editing of an mxf file
// This is a generic code that can be reused with any mxf file
void display_output_material(generic_list*);

// Displays the embedded essence of an mxf file
// This is a generic code that can be reused with any mxf file
void display_embedded_material(concrete_list*, std::string);

// essence extractor -> write audiovisual material on the disk
void read_essence(I_concrete_track*, std::string);

int main(int argc, char* argv[])
{
  // You should provide all the opAtom files that need to be synchronize
  // in order to rebuild the complete output.
  // If some of the files are missing it is still possible to manipulate
  // the files however the output will be incomplete.
  if (argc < 2)
  {
    wcout << L"usage : in_1.mxf ... in_n.mxf (opatom mxf files)" << endl;
    exit(0);
  }

  mxf_file_list* files = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    wcout << L"Build list of opatom files" << endl;
    GetEmptyMxfFileListInterface(&files);
    int i;
    for (i=1; i < argc; i++)
    {
      wcout << L"Loading opatom file " << argv[i] << endl;
      I_mxf_file* f = NULL;
      NewMxfFileInterface(&f, argv[i]);
      files->add(&f);
    }
    throw_on_error();

    wcout << L"Synchronize files" << endl;
    I_opatom_assembler* pi_assembler = NULL;
    NewOpAtomAssemblerInterface(&pi_assembler);
    // This function links the OpAtom files together thanks to their UMID
    pi_assembler->synchronize(files);
    FreeOpAtomAssemblerInterface(&pi_assembler);
    throw_on_error();

    // Following steps are only for display purpose

    // Note
    // after this sync step, each opatom file share the same generic material,
    // and therefore from any file it becomes possible to access the embedded
    // essence from other OpAtom files.
    // Here we cycle through the Opatom files to demonstrate they all have the same
    // generic material however the whole output could be retrieved from the genric
    // material of a single file.

    // Cycle through OpAtom files.
    files->begin();
    I_mxf_file** p_f = NULL;
    i=0;
    while ((p_f = files->next()) != NULL)
    {
      wcout << L"*************************************************************" << endl;
      wcout << L"Display information for " << argv[i+1] << L" file" << endl;
      i++;
      I_mxf_file* pi_mxf_file = *p_f;

      const I_metadata* preface = pi_mxf_file->get_preface();
      if (!preface)
      {
	wcout << L"Error while loading preface" << endl;
	return 0;
      }

      // Read MXF file version
      read_file_version(preface);
      read_history(pi_mxf_file);

      std::string op_pattern = pi_mxf_file->operational_pattern();
      wcout << L"Operational Pattern: " << op_pattern.c_str() << endl;
  	
      generic_list* output_mat = pi_mxf_file->output_material();
      wcout << endl << L"Outputs (Material Package): " << (long)output_mat->size() << endl;

      concrete_list* embedded_mat = pi_mxf_file->embedded_material();
      wcout << L"Source Clips (Top-Level Source Package): " << (long)embedded_mat->size() << endl;

      // Rarely used
      generic_list* low_level_mat = pi_mxf_file->low_level_material();
      wcout << L"Low Level Clips (Low-Level Source Package): " << (long)low_level_mat->size() << endl;	

      display_output_material(output_mat);
      throw_on_error();

      display_embedded_material(embedded_mat, std::string(argv[i]));
      throw_on_error();

      pi_mxf_file->free_generic_list(output_mat);
      pi_mxf_file->free_concrete_list(embedded_mat);
      pi_mxf_file->free_generic_list(low_level_mat);
    }
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  if (files)
  {
    files->begin();
    I_mxf_file** p_f = NULL;
    p_f = files->next();
    if (p_f!=NULL)
    {
      // because the files were synchronized, calling FreeMxfFileInterface
      // on one of them will destroy all of them.
      I_mxf_file* pi_mxf_file = *p_f;
      FreeMxfFileInterface(&pi_mxf_file);
    }
    FreeMxfFileListInterface(&files);
  }

  return 0;
}

void read_file_version(const I_metadata* preface)
{
  properties_list* vect = preface->get_properties();
  vect->begin();
  I_property** prop_p = NULL;
  while ((prop_p = vect->next())!=NULL)
  {
    I_property* prop = *prop_p;
    std::string name = prop->get_label();
    const I_value* val = prop->get_value();
    const uint8_t* data = val->get_data();

    if (name==APPLICATION_VERSION)
    {
      uint16_t version = *((uint16_t*)data);
      wstringstream str;
      short major = version/256;
      str << L"MXF v" << (short) major << L"." << (short) (version-major*256) << endl; 
      wcout << str.str();
      break;
    }
  }
  preface->free_properties(vect);
}

void read_history(I_mxf_file* pi_mxf_file)
{
  wcout << L"History of modifications :" << endl;
  metadata_list* idents=pi_mxf_file->get_identifications();
  idents->begin();
  I_metadata** ident_p = NULL;
  while ((ident_p = idents->next())!=NULL)
  {
    const I_metadata* ident = *ident_p;
	
    std::wstring date_str = L"";
    std::wstring companyname = L"";
    std::wstring productname = L"";
    std::wstring productversion = L"";

    properties_list* vect = ident->get_properties();
    vect->begin();
    I_property** prop_p = NULL;
    while ((prop_p = vect->next())!=NULL)
    {
      I_property* prop = *prop_p;
      std::string name = prop->get_label();
      const I_value* val = prop->get_value();
      const uint8_t* data = val->get_data();
		
      if (name==MODIFICATION_DATE)
      {
	uint16_t* date = (uint16_t*)data;
      				
	wstringstream str;
	str << "Modified on " << (short) date[1] << "/" << (short) date[2] 
	    << "/" << (short) date[0] << " at " << (short) date[3] 
	    << ":" << (short) date[4] << ":" << (short) date[5]; 
	date_str = str.str();
      }
      else if (name==APPLICATION_SUPPLIER)
      {
	companyname = std::wstring((wchar_t*)data);
      }
      else if (name==APPLICATION_NAME)
      {
	productname = std::wstring((wchar_t*)data);
      }
      else if (name==APPLICATION_VERSION_STRING)
      {
	productversion = std::wstring((wchar_t*)data);
      }
    }
    
    ident->free_properties(vect);	
    wcout << date_str.c_str() << L" by " << companyname.c_str() << L" " << productname.c_str() << L" " << productversion.c_str() << endl;
  }
  pi_mxf_file->free_identifications(idents);

}

void display_output_material(generic_list* output_mat)
{
  wcout << endl << L"Output Material :" << endl;
  int id_s = 0;
  output_mat->begin();
  I_generic_material** pi_mat_p = NULL;
  while ((pi_mat_p = output_mat->next())!=NULL)
  {
    I_generic_material* pi_mat = *pi_mat_p;

    size_t s = size_t(0);
    const wchar_t* name_m = pi_mat->get_material_name(s);
    if (name_m)
    {
      wcout << L"Package " << id_s << L": " << name_m << endl;
    }
    else
    {
      wcout << L"Package " << id_s << endl;
    }

    // Retrieve timecode information for this material
    wcout << L"Timecode :" << endl;
    I_timecode_material* tm_mat = pi_mat->get_timecode();
    if (tm_mat)
    {
      rational *editrate = tm_mat->get_edit_rate();
      int32_t ed = editrate->get_num()/editrate->get_den();
      FreeRationalInterface(&editrate);
      int64_t origin = tm_mat->get_origin();
      int64_t duration = tm_mat->get_total_duration();

      wstringstream str_time;
      str_time << L"\tTimecode track starts at " << (long) origin << L"eu lasts " << (long)duration << L"eu @" << (int)ed << L"eu/s" << endl;
      wcout << str_time.str();

      // Loop through timecode components
      ordered_timecode_list* vect_time = tm_mat->get_timecode_list();
      vect_time->begin();
      I_timecode_component** tmc_p = NULL;
      while ((tmc_p = vect_time->next())!=NULL)
      {
	wstringstream tmstr;
	I_timecode_component* tmc = *tmc_p;
	tmstr << L"\t" << (tmc->get_timecode())->get_time() << L" lasts " << (long)tmc->get_duration() << L" eu" << endl;
	wcout << tmstr.str();
      }
      tm_mat->free_timecode_list(vect_time);
    }

    // Cycle through essence tracks
    wcout << "\tTracks: " << endl;
    track_list* source_tr = pi_mat->source_tracks();
    source_tr->begin();
    I_track** trp = NULL;
    while ((trp = source_tr->next())!=NULL)
    {
      I_track* tr = *trp;
      wstringstream str;
      rational *r = tr->edit_rate();
      double rate = r->get_double();
      FreeRationalInterface(&r);
      str << L"\t\t[ " << tr->type_name() << L" : start " << (long)tr->origin() << L" lasts " << (long) tr->duration() << L"eu @" << (int) rate << L"eu/s" << L" ]" << endl;
      wcout << str.str();
      ordered_track_item_list* sc = tr->elements();
      sc->begin();
      I_track_item** s_p = NULL;
      while ((s_p = sc->next())!=NULL)
      {
	wstringstream str3;
	I_track_item* s2 = *s_p;
	I_source_clip* sc2 = source_clip_cast(s2);
	str3 << L"\t\t\tusing sub_clip from " << (long)sc2->get_start_position() << L" for " << (long)sc2->get_duration() << L"eu" << endl;
	wcout << str3.str();
      }
      tr->free_elements(sc);
    }
      
    pi_mat->free_tracks(source_tr);
		
    id_s++;
    cout << endl;
  }
}

void display_embedded_material(concrete_list* embedded_mat, std::string name)
{
  wcout << endl << L"Embedded Material :" << endl;
  // loop through embedded material.
  embedded_mat->begin();
  I_concrete_material** pi_mat_p = NULL;
  while ((pi_mat_p = embedded_mat->next())!=NULL)
  {
    I_concrete_material* pi_conc = *pi_mat_p;

    size_t s;
    const wchar_t* name_m = pi_conc->get_material_name(s);
    if (name_m)
      wcout << L"Package " << name_m << endl;
    else
      wcout << L"New Package " << endl;

    // Retrieve timecode information for this material
    wcout << L"Timecode :" << endl;
    I_timecode_material* tm_mat = pi_conc->get_timecode();
    if (tm_mat)
    {
      rational *editrate = tm_mat->get_edit_rate();
      double ed = editrate->get_double();	   
      int64_t origin = tm_mat->get_origin();
      int64_t duration = tm_mat->get_total_duration();

      wstringstream str_time;
      str_time << L"\tTimecode starts at " << (long) origin << L"eu lasts " << (long)duration << L"eu @" << (double)ed << L"eu/s" << endl;
      wcout << str_time.str();

      // Retrieve list of timecode components
      ordered_timecode_list* vect_time = tm_mat->get_timecode_list();
      vect_time->begin();
      I_timecode_component** time_p = NULL;
      while ((time_p = vect_time->next())!=NULL)
      {
	wstringstream tmstr;
	I_timecode_component* tmc = *time_p;
	tmstr << L"\t" << (tmc->get_timecode())->get_time() << L" lasts " << (long)tmc->get_duration() << L" eu" << endl;
	wcout << tmstr.str();
      }
      tm_mat->free_timecode_list(vect_time);
    }

    // display universal identifier UMID of this material
    I_umid* um = pi_conc->get_umid();
    wcout << L"\tUMID " << um->get_str() << endl;
    FreeUmidInterface(&um);
		
    // loop through data tracks
    wcout << L"\tTracks: " << endl;
    track_list* source_tr = pi_conc->source_tracks();
    source_tr->begin();
    I_track** trp = NULL;
    while ((trp = source_tr->next())!=NULL)
    {
      I_concrete_track* ctr = (I_concrete_track*)*trp;
      wstringstream str;

      rational *r = ctr->edit_rate();
      double rate = r->get_double();
      FreeRationalInterface(&r);
      str << L"\t\t[ " << ctr->type_name() << L" : start " << (long)ctr->origin() << L" lasts " << (long) ctr->duration() << L"eu @" << (double) rate << L"eu/s" << L" ]" << endl;
      wcout << str.str();

      // Extract essence in a file
      read_essence(ctr, name);		
    }

    pi_conc->free_tracks(source_tr);
    wcout << endl;
  }
}

void read_essence(I_concrete_track* tr, std::string name)
{
  wcout << L"\t\tRetrieve information on the track's content" << endl;
  I_essence_type* ess_type = tr->get_essence_type();
  throw_on_error();
  if (!ess_type)
    throw 1;

  // Retrieve essence type to write with correct extension
  std::string file_extension = "unknown";
  switch (ess_type->get_type())
  {     
  case ess_d10:
  {
    if (tr->type()==sound)
    {
      wcout << L"\t\tD10 Audio AES3" << endl; 
      file_extension = ".aes";
      break;
    }
    else if (tr->type()==picture)
    {
      wcout << L"\t\tD10 Video IMX" << endl;
      file_extension = ".imx";
      break;
    }
  }
  case ess_dv_iec:                 
  case ess_dv_cam_iec:             
  {
    wcout << L"\t\tDV IEV" << endl;
    file_extension = ".dv";
    break;
  }
  case ess_dv_smpte:               
  {
    wcout << L"\t\tDV SMPTE" << endl;
    file_extension = ".dif";
    break;
  }
  case ess_mpeg_es:
  {
    wcout << L"\t\tMPEG Elementary Stream" << endl;
    file_extension = ".m2v";
    break;
  }
  case ess_mpeg_pes:
  {
    wcout << L"\t\tMPEG Packetized Elementary Stream" << endl;
    file_extension = ".mpg";
    break;
  }
  case ess_mpeg_ps:
  {
    wcout << L"\t\tMPEG Program Stream" << endl;
    file_extension = ".mpg";
    break;
  }
  case ess_mpeg_ts:                  
  {
    wcout << L"\t\tMPEG Transport Stream" << endl;
    file_extension = ".mpg";
    break;
  }
  case ess_bwf:
  {
    wcout << L"\t\tBroadcast Wave" << endl;
    file_extension = ".wav";
    break;
  }
  case ess_aes3: 
  {
    wcout << L"\t\tAudio AES3" << endl;
    file_extension = ".aes";
    break;
  }
  default:
    break;
  }

  filebuf ofs;
  stringstream s_id;
  s_id << (long)tr->get_track_id();
  string str_id = "_" + s_id.str();

  ofs.open( (name + str_id + file_extension).c_str(), ios::out | ios::binary  );

  // get track's origin
  int64_t ori = tr->origin();

  wcout << L"\t\tSeek at the beginning of the track" << endl;
  tr->seek_eu(-ori);

  // loop as long as some data is ready to be read.
  wcout << L"\t\tWriting essence file edit unit by edit unit" << endl;
  uint64_t eu_size= 0;
  while ((eu_size = tr->current_eu_size())!=0)
  {
    uint8_t* buffer = new uint8_t[size_t(eu_size)];
    size_t read = tr->read_eu(buffer, size_t(eu_size));
    ofs.sputn((char*) &buffer[0], std::streamsize(read) );
    delete [] buffer;
    tr->seek_next_eu();
    throw_on_error();
  }

  ofs.close();
}


void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
